"""modified from https://github.com/nttcslab/byol-a """
"""BYOL for Audio: Downstream task dataset utility.

The `create_data_source()` will generate the data source class object for downstream tasks.
The data source will provide task training details such as labels, fold splits,
classes, as well as audio file paths.
"""

import os
import sys
import shutil
import random
import numpy as np
from pathlib import Path
import pandas as pd
import copy
import re
import logging
from collections import defaultdict
from itertools import chain
import subprocess
import torch
from torch.utils.data import Dataset
import torchaudio
from audiossl.datasets import register_dataset




def read_task_df(task, base_folder):
    curdir = os.path.dirname(__file__)
    df = pd.read_csv(Path(os.path.join(curdir,"byol_a_meta"))/f'{task}.csv')
    # replace all str label with int label
    df.label = df.label.map({l: i for i, l in enumerate(df.label.unique())})
    return df


def get_us8k(base_folder):
    df = read_task_df('us8k', base_folder)
    # add fold column
    df['fold'] = df.file_name.map(lambda s: int(s.split('/')[1][4:])) # 'audio/foldXX/*.wav'
    idxs = [None for _ in range(10)]
    for i, sdf in df.groupby('fold'):
        idxs[i - 1] = sdf.index.values
        print(f' {i}:{len(sdf)}', end='')
    print(' samples.')
    return df, idxs, True


def get_spc(base_folder, ver):
    df = read_task_df(f'spcv{ver}', base_folder)
    # make split index
    idxs = [None, None, None]
    idxs[0] = df[df.split == 'train'].index.values
    idxs[1] = df[df.split == 'val'].index.values
    idxs[2] = df[df.split == 'test'].index.values
    assert np.all(np.array([len(idx) for idx in idxs]) > 0)
    return df, idxs, False


def get_spcv1(base_folder):
    return get_spc(base_folder, 1)


def get_spcv2(base_folder):
    return get_spc(base_folder, 2)


def get_nsynth(base_folder):
    df = read_task_df('nsynth', base_folder)
    # make split index
    idxs = [None, None, None]
    idxs[0] = df[df.split == 'train'].index.values
    idxs[1] = df[df.split == 'valid'].index.values
    idxs[2] = df[df.split == 'test'].index.values
    assert np.all(np.array([len(idx) for idx in idxs]) > 0)
    return df, idxs, False


def get_fsdnoisy18k(base_folder):
    df = read_task_df('fsdnoisy18k', base_folder)
    # make split index
    idxs = [None, None, None]
    idxs[0] = df[df.split == 'train'].index.values
    idxs[1] = df[df.split == 'valid'].index.values
    idxs[2] = df[df.split == 'test'].index.values
    assert np.all(np.array([len(idx) for idx in idxs]) > 0)
    return df, idxs, False


def load_metadata(mode, base):
    getter = eval('get_' + mode)
    df, fold_idxs, loocv = getter(base)
    return df, fold_idxs, loocv


class BaseDataSource:
    """Data source base class, see TaskDataSource for the detail."""

    def __init__(self, df, fold_idxes, loocv):
        self.df, self.fold_idxes, self.loocv = df, fold_idxes, loocv
        self.get_idxes = None

    @property
    def labels(self):
        if self.get_idxes is not None:
            return self.df.label.values[self.get_idxes]
        return self.df.label.values

    def __len__(self):
        if self.get_idxes is None:
            return len(self.df)
        return len(self.get_idxes)

    def index_of_folds(self, folds):
        idxes = []
        for fold in folds:
            idxes.extend(self.fold_idxes[fold])
        return idxes

    def subset_by_idxes(self, idxes):
        dup = copy.copy(self)
        dup.get_idxes = idxes
        return dup

    def subset(self, folds):
        """Returns a subset data source for the fold indexes.
        folds: List of fold indexes.
        """
        return self.subset_by_idxes(self.index_of_folds(folds))

    @property
    def n_folds(self):
        return len(self.fold_idxes)

    @property
    def n_classes(self):
        return len(set(self.df.label.values))

    def real_index(self, index):
        if self.get_idxes is not None:
            index = self.get_idxes[index]
        return index


class TaskDataSource(BaseDataSource):
    """Downstream task data source class.

    This class provides files and metadata in the dataset,
    as well as methods to manage data splits/folds.

    Properties:
        files: Audio sample pathnames.
        labels: List of int, class indexes of samples.
        classes: List of int, possible class indexes. [0, 1, 2, ] for example of 3 classes.
        n_classes: Number of classes
        n_folds: Number of folds.

    Methods:
        subset: Makes subset data source access.
    """

    def __init__(self, audio_folder, meta_folder, mode):
        super().__init__(*load_metadata(mode, meta_folder))
        self.audio_folder = Path(audio_folder + mode)

    def file_name(self, index):
        index = self.real_index(index)
        return self.audio_folder/self.df.file_name.values[index]

    @property
    def files(self):
        return [self.file_name(i) for i in range(len(self))]


def create_data_source(mode):
    """Creates data source object for downstream task you want."""

    assert mode in ['us8k', 'spcv1', 'spcv2', 'nsynth', 'fsdnoisy18k']
    return TaskDataSource(mode)

class Nsynth(Dataset):
    def __init__(self,path,split="train",sr=16000,transform=None):
        df = read_task_df('nsynth', path)
        indexs = df[df.split == split].index.values
        self.file_names=df.file_name.values[indexs]
        self.labels=df.label.values[indexs]
        self.num_classes = len(set(df.label.values))
        self.path = path
        self.sr=sr
        self.transform=transform
    def __getitem__(self,index):
        file_name = self.file_names[index]
        label = self.labels[index]
        waveform, sr = torchaudio.load(os.path.join(self.path,file_name),normalize=True)
        if not (self.sr == sr):
            raise "sampling rate {} is expected, while {} is given".format(
                self.sr, sr)
        if self.transform is not None:
            return self.transform(waveform),label
        else:
            return waveform,label

    def __len__(self):
        return len(self.file_names)

class Urbansound8k(Dataset):
    def __init__(self,path,split="train",valid_fold=10,sr=16000,transform=None):

        df = read_task_df('us8k', path)
        # add fold column
        df['fold'] = df.file_name.map(lambda s: int(s.split('/')[1][4:])) # 'audio/foldXX/*.wav'
        idxs = [None for _ in range(10)]
        for i, sdf in df.groupby('fold'):
            idxs[i - 1] = sdf.index.values
        indexs = []
        if split=="train":
            for i in range(10):
                if i==valid_fold-1:
                    continue
                indexs.extend(idxs[i])
        else:
            indexs.extend(idxs[valid_fold-1])

        self.file_names=df.file_name.values[indexs]
        self.labels=df.label.values[indexs]
        self.num_classes = len(set(df.label.values))
        self.path = path
        self.sr=sr
        self.transform=transform
    def __getitem__(self,index):
        file_name = self.file_names[index]
        label = self.labels[index]
        waveform, sr = torchaudio.load(os.path.join(self.path,file_name),normalize=True)
        if not (self.sr == sr):
            raise "sampling rate {} is expected, while {} is given".format(
                self.sr, sr)
        if self.transform is not None:
            return self.transform(waveform),label
        else:
            return waveform,label

    def __len__(self):
        return len(self.file_names)




from pathlib import Path
import os

from .registry import register_dataset,list_all_datasets,get_dataset

from .lmdb import LMDBDataset
from .byol_a import Nsynth,Urbansound8k
from .voxceleb1 import SpeakerClassifiDataset
from .librispeech import LibriSpeechDataset
from .iemocap import IEMOCAPDataset

@register_dataset("voxceleb1",multi_label=False,num_labels=1251,num_folds=1)
def create_voxceleb1(data_path,split,transform,target_transform):
    if split == "valid":
        split = "dev"
    return SpeakerClassifiDataset(split,
                                Path(data_path),
                                Path(os.path.join(data_path,"iden_split.txt")),
                                transform=transform,
                                target_transform=target_transform)
@register_dataset("us8k",multi_label=False,num_labels=10,num_folds=10)
def create_us8k(data_path,split,fold,transform,target_transform):
    return Urbansound8k(data_path,split=split,valid_fold=fold,transform=transform)
    
@register_dataset("nsynth",multi_label=False,num_labels=11,num_folds=1)
def create_nsynth(data_path,split,transform,target_transform):
    return Nsynth(data_path,split=split,transform=transform)

@register_dataset("spcv2",multi_label=False,num_labels=35,num_folds=1)
def create_spcv2(data_path,split,transform,target_transform):
    if split == "test":
        split = "eval"
    return LMDBDataset(data_path,split=split,transform=transform,target_transform=target_transform) 

@register_dataset("audioset_b",multi_label=True,num_labels=527,num_folds=1)
def create_spcv2(data_path,split,transform,target_transform):
    if split == "test":
        split = "eval"
    return LMDBDataset(data_path,split=split,transform=transform,target_transform=target_transform) 

@register_dataset("audioset",multi_label=True,num_labels=527,num_folds=1)
def create_spcv2(data_path,split,transform,target_transform):
    if split == "test":
        split = "eval"
    return LMDBDataset(data_path,split=split,transform=transform,target_transform=target_transform) 



__all__ = ['LMDBDataset',
           'Nsynth',
           'Urbansound8k',
           'LibriSpeechDataset',
           'SpeakerClassifiDataset',
           'IEMOCAPDataset']



import json
import pprint
DATASET_REGISTRY={}

class DatasetInfo:
    """Placeholder for properties of dataset
    """
    def __init__(self, **kwargs): 
        self.__dict__.update(kwargs)
    def __str__(self):
        pp = pprint.PrettyPrinter(indent=2)
        return pp.pformat(self.__dict__)


def get_dataset(name:str) -> DatasetInfo:
    """
    Example::

            datasetinfo = get_dataset("spcv2")
            print(datasetinfo)
            creat_fn=datasetinfo.creator
            num_labels=datsetinfo.num_labels
            multi_label=datasetinfo.multi_label

    """
    if name in DATASET_REGISTRY.keys():
        return DATASET_REGISTRY[name]
    else:
        raise RuntimeError("dataset {} is not registered".format(name))

def list_all_datasets():
    for k,v in DATASET_REGISTRY.items():
        print("{}:\n{}".format(k,v))



def register_dataset(name,**kwargs):
    info = dict(**kwargs)

    def register_dataset_(creator):
        info.update({"creator":creator})
        if name in DATASET_REGISTRY.keys() and not (get_dataset(name).creator == creator):
            raise RuntimeError("dataset {} has been already registered".format(name))
        DATASET_REGISTRY.update({name:DatasetInfo(**info)})
        return creator
    return register_dataset_

def add_regist(name,**kwargs):
    info = dict(**kwargs)
    if name in DATASET_REGISTRY.keys() and not (get_dataset(name).creator == info["creator"]):
        raise RuntimeError("dataset {} has been already registered".format(name))
    DATASET_REGISTRY.update({name:DatasetInfo(**info)})


from tkinter import W
import pytorch_lightning


from pytorch_lightning import Trainer
from pytorch_lightning import LightningDataModule,LightningModule

class EmbeddingExtractor:
    def __init__(self,
                 module:LightningModule,
                 nproc:int=1
                ):
        self.trainer = Trainer(
                            strategy="ddp_find_unused_parameters_false",
                            sync_batchnorm=True,
                            logger=False,
                            gpus=nproc,
                            #profiler="simple",
                            #max_epochs=1,
                            )
        self.module = module
    def extract(self,dataloader):
        return self.trainer.predict(self.module,dataloader)






import torchaudio
import sys
from torch.utils.data import ConcatDataset



class LibriSpeechDataset:
    def __init__(self,path,transform=None):
        ds1=torchaudio.datasets.LIBRISPEECH(path,url="train-clean-100",download=False)
        ds2=torchaudio.datasets.LIBRISPEECH(path,url="train-clean-360",download=False)
        ds3=torchaudio.datasets.LIBRISPEECH(path,url="train-other-500",download=False)
        self._ds = ConcatDataset([ds1,ds2,ds3])
        self.transform=transform
    def __getitem__(self,index):
        item = self._ds[index]
        wav,sr = item[0],item[1]
        if self.transform is not None:
            return self.transform(wav),0
        else:
            return wav,0
    def __len__(self):
        return len(self._ds)




from pytorch_lightning import LightningDataModule
import torch
from torch.utils import data
from audiossl import datasets

def get_inmemory_datamodule(x_train,
                            y_train,
                            x_val,
                            y_val,
                            x_test,
                            y_test,
                            batch_size):
            
    dataset_train = data.TensorDataset(x_train,y_train)
    dataset_val = data.TensorDataset(x_val,y_val)
    dataset_test = data.TensorDataset(x_test,y_test)

    return LightningDataModule.from_datasets(dataset_train,dataset_val,dataset_test,batch_size=batch_size)

class DownstreamDataModule(LightningDataModule):
    def __init__(self,
                 data_path:str,
                 dataset_name:str,
                 fold:int = 0,
                 batch_size_per_gpu=1024,
                 num_workers=10,
                 transforms=[None,None,None],
                 target_transforms=[None,None,None],
                 collate_fn=None,
                 limit_batch_size=None,
                 shuffle = True,
                 **kwargs
                 ):
        super().__init__()
        self.batch_size=min(limit_batch_size,batch_size_per_gpu)\
            if limit_batch_size is not None else batch_size_per_gpu
        self.num_workers=num_workers
        self.transforms=transforms
        self.target_transforms=target_transforms
        self.collate_fn = collate_fn
        dataset_info=datasets.get_dataset(dataset_name)
        num_folds = dataset_info.num_folds
        self.num_labels=dataset_info.num_labels
        self.multi_label=dataset_info.multi_label
        self.shuffle = shuffle
        if num_folds > 1:
            self.dataset_train = dataset_info.creator(data_path,
                                                      "train",
                                                      fold,
                                                      transforms[0],
                                                      target_transform=target_transforms[0])
            self.dataset_val = dataset_info.creator(data_path,
                                                      "valid",
                                                      fold,
                                                      transforms[1],
                                                      target_transform=target_transforms[1])
            self.dataset_test = dataset_info.creator(data_path,
                                                      "test",
                                                      fold,
                                                      transforms[2],
                                                      target_transform=target_transforms[2])
        else:
            self.dataset_train = dataset_info.creator(data_path,
                                                      "train",
                                                      transforms[0],
                                                      target_transform=target_transforms[0])
            self.dataset_val = dataset_info.creator(data_path,
                                                      "valid",
                                                      transforms[1],
                                                      target_transform=target_transforms[1])
            self.dataset_test = dataset_info.creator(data_path,
                                                      "test",
                                                      transforms[2],
                                                      target_transform=target_transforms[2])
        self.save_hyperparameters()
    def prepare_data(self):
        pass

    def train_dataloader(self):
        return data.DataLoader(self.dataset_train,
                        batch_size=self.batch_size,
                        num_workers=self.num_workers,
                        shuffle=self.shuffle,
                        sampler=None,
                        drop_last=False,
                        collate_fn=self.collate_fn,
                        pin_memory=True)
    def val_dataloader(self):
        return data.DataLoader(self.dataset_val,
                        batch_size=self.batch_size,
                        num_workers=self.num_workers,
                        shuffle=False,
                        sampler=None,
                        drop_last=False,
                        collate_fn=self.collate_fn,
                        pin_memory=True)
    def test_dataloader(self):
        return data.DataLoader(self.dataset_test,
                        batch_size=self.batch_size,
                        num_workers=self.num_workers,
                        shuffle=False,
                        sampler=None,
                        drop_last=False,
                        collate_fn=self.collate_fn,
                        pin_memory=True)


    @staticmethod
    def add_data_specific_args(parent_parser):
        parser = parent_parser.add_argument_group("DownstreamData")
        parser.add_argument("--dataset_name", type=str, default=None, help="dataset name")
        parser.add_argument("--data_path", type=str, default=None, help="data path")
        parser.add_argument('--batch_size_per_gpu', default=256, type=int,
            help='Per-GPU batch-size : number of distinct samples loaded on one GPU.')
        parser.add_argument('--num_workers', default=10, type=int, help='Number of data loading workers per GPU.')
        return parent_parser



from torchvision import transforms 
import torchaudio
import torch
import numpy as np
import time
from torch.nn import functional as F


class CustomAudioTransform:
    def __repr__(self):
        return self.__class__.__name__ + '()'
class Identity(CustomAudioTransform):
    def __call__(self,signal):
        return signal
    
class GaussianNoise(CustomAudioTransform):
    def __init__(self,g):
        self.g = g
    def __call__(self,signal):
        return signal + self.g * torch.randn_like(signal)

class PadToSize(CustomAudioTransform):
    def __init__(self, size:int):
        self.size = size

    def __call__(self, signal):

        if signal.shape[1] < self.size :
            signal = F.pad(signal, (0, self.size-signal.shape[1]))
        return signal

class ToSizeN(CustomAudioTransform):
    def __init__(self, size:int):
        self.size = size

    def __call__(self, signal):
        n = signal.shape[1]//self.size
        m = signal.shape[1] % self.size
        if m > self.size//2 or n==0:
            signal = F.pad(signal, (0, self.size*(n+1)-signal.shape[1]))
        else:
            signal = F.pad(signal, (0, self.size*n-signal.shape[1]))
        return signal

class CentralCrop(CustomAudioTransform):
    def __init__(self, size:int, pad:bool=True):
        self.size = size
        self.pad = pad

    def __call__(self, signal):

        if signal.shape[1] < self.size :
            if self.pad:
                signal = F.pad(signal, (0, self.size-signal.shape[1]))
            return signal

        start = (signal.shape[1] - self.size) // 2
        return signal[:, start: start + self.size]

class RandomCrop(CustomAudioTransform):
    def __init__(self, size:int, pad:bool = True):
        self.size = size
        self.pad = pad

    def __call__(self, signal):
        if signal.shape[1] < self.size :
            if self.pad:
                signal = F.pad(signal, (0, self.size-signal.shape[1]))
            return signal
        start = np.random.randint(0, signal.shape[-1] - self.size + 1)
        return signal[:, start: start + self.size]
    

class Normalize(CustomAudioTransform):
    def __init__(self,std_mean=None,reduce_dim=None):
        self.std_mean = std_mean
        self.reduce_dim = reduce_dim
    def __call__(self,input):
        """
        assuming input has shape [batch,nmels,time]
        """
        std,mean = None,None
        if self.std_mean is None:
            if self.reduce_dim is not None:
                std,mean = torch.std_mean(input,dim=self.reduce_dim,keepdim=True)
            else:
                std,mean = torch.std_mean(input)
        else:
            std,mean = self.std_mean
        output = input - mean 
        output = output / (std + 1e-6)
        return output

class MinMax(CustomAudioTransform):
    def __init__(self,min,max):
        self.min=min
        self.max=max
    def __call__(self,input):
        min_,max_ = None,None
        if self.min is None:
            min_ = torch.min(input)
            max_ = torch.max(input)
        else:
            min_ = self.min
            max_ = self.max
        input = (input - min_)/(max_- min_) *2. - 1.
        return input

class div(CustomAudioTransform):
    def __init__(self,value=100):
        self.value = value
    def __call__(self,input):
        input /= 100
        return input



# copied and modified from https://github.com/nttcslab/byol-a
import random 
import numpy as np
import torch
from torch import nn
from torch.nn import functional as F
class RandomResizeCrop(nn.Module):
    """Random Resize Crop block.

    Args:
        virtual_crop_scale: Virtual crop area `(F ratio, T ratio)` in ratio to input size.
        freq_scale: Random frequency range `(min, max)`.
        time_scale: Random time frame range `(min, max)`.
    """

    def __init__(self, virtual_crop_scale=(1.0, 1.5), freq_scale=(0.6, 1.5), time_scale=(0.6, 1.5)):
        super().__init__()
        self.virtual_crop_scale = virtual_crop_scale
        self.freq_scale = freq_scale
        self.time_scale = time_scale
        self.interpolation = 'bicubic'
        assert time_scale[1] >= 1.0 and freq_scale[1] >= 1.0

    @staticmethod
    def get_params(virtual_crop_size, in_size, time_scale, freq_scale):
        canvas_h, canvas_w = virtual_crop_size
        src_h, src_w = in_size
        h = np.clip(int(np.random.uniform(*freq_scale) * src_h), 1, canvas_h)
        w = np.clip(int(np.random.uniform(*time_scale) * src_w), 1, canvas_w)
        i = random.randint(0, canvas_h - h) if canvas_h > h else 0
        j = random.randint(0, canvas_w - w) if canvas_w > w else 0
        return i, j, h, w

    def forward(self, lms):
        # make virtual_crop_arear empty space (virtual crop area) and copy the input log mel spectrogram to th the center
        virtual_crop_size = [int(s * c) for s, c in zip(lms.shape[-2:], self.virtual_crop_scale)]
        virtual_crop_area = (torch.zeros((lms.shape[0], virtual_crop_size[0], virtual_crop_size[1]))
                             .to(torch.float).to(lms.device))
        _, lh, lw = virtual_crop_area.shape
        c, h, w = lms.shape
        x, y = (lw - w) // 2, (lh - h) // 2
        virtual_crop_area[:, y:y+h, x:x+w] = lms
        # get random area
        i, j, h, w = self.get_params(virtual_crop_area.shape[-2:], lms.shape[-2:], self.time_scale, self.freq_scale)
        crop = virtual_crop_area[:, i:i+h, j:j+w]
        # print(f'shapes {virtual_crop_area.shape} {crop.shape} -> {lms.shape}')
        lms = F.interpolate(crop.unsqueeze(0), size=lms.shape[-2:],
            mode=self.interpolation, align_corners=True).squeeze(0)
        return lms.to(torch.float)

    def __repr__(self):
        format_string = self.__class__.__name__ + f'(virtual_crop_size={self.virtual_crop_scale}'
        format_string += ', time_scale={0}'.format(tuple(round(s, 4) for s in self.time_scale))
        format_string += ', freq_scale={0})'.format(tuple(round(r, 4) for r in self.freq_scale))
        return format_string

def energy_scale(xa,xb):
    xa=xa.exp()
    xb=xb.exp()
    return torch.sum(xa)/torch.sum(xb)
def log_mixup_exp(xa, xb, alpha):
    xa = xa.exp()
    xb = xb.exp()
    len_xa = xa.shape[2]
    len_xb = xb.shape[2]
    if len_xa < len_xb:
        start = np.random.randint(0,len_xb-len_xa)
        scale = energy_scale(xa,xb[:,:,start:start+len_xa])
        scale = 1.0
        xa = alpha*xa +(1.-alpha)*scale*xb[:,:,start:start+len_xa]
        return torch.log(xa + torch.finfo(xa.dtype).eps)
    elif len_xa > len_xb:
        start = np.random.randint(0,len_xa-len_xb)
        scale = energy_scale(xa[:,:,start:start+len_xb],xb)
        scale = 1.0
        xa[:,:,start:start+len_xb] = alpha*xa[:,:,start:start+len_xb] +(1.-alpha)*scale*xb
        return torch.log(xa + torch.finfo(xa.dtype).eps)
    else:
        scale = energy_scale(xa,xb)
        scale = 1.0
        x =  alpha * xa + (1.-alpha)*scale*xb
        return torch.log(x + torch.finfo(x.dtype).eps)


class Mixup(nn.Module):
    """Mixup.

    Args:
        ratio: Alpha in the paper.
        n_memory: Size of memory bank FIFO.
        log_mixup_exp: Use log-mixup-exp to mix if this is True, or mix without notion of log-scale.
    """

    def __init__(self, ratio=0.4, n_memory=2000, log_mixup_exp=True):
        super().__init__()
        self.ratio = ratio
        self.n = n_memory
        self.log_mixup_exp = log_mixup_exp
        self.memory_bank = []

    def forward(self, x):
        # mix random
        alpha = self.ratio * np.random.random()
        if self.memory_bank:
            # get z as a mixing background sound
            z = self.memory_bank[np.random.randint(len(self.memory_bank))]
            # mix them
            mixed = log_mixup_exp(x, z, 1. - alpha) if self.log_mixup_exp \
                    else alpha * z + (1. - alpha) * x
        else:
            mixed = x
        # update memory bank
        self.memory_bank = (self.memory_bank + [x])[-self.n:]

        return mixed.to(torch.float)

    def __repr__(self):
        format_string = self.__class__.__name__ + f'(ratio={self.ratio},n={self.n}'
        format_string += f',log_mixup_exp={self.log_mixup_exp})'
        return format_string


class MixGaussianNoise():
    """Gaussian Noise Mixer.
    This interpolates with random sample, unlike Mixup.
    """

    def __init__(self, ratio=0.3):
        self.ratio = ratio

    def forward(self, lms):
        x = lms.exp()

        lambd = self.ratio * np.random.rand()
        z = torch.normal(0, lambd, x.shape).exp()
        mixed = (1 - lambd) * x + z + torch.finfo(x.dtype).eps

        return mixed.log()

    def __repr__(self):
        format_string = self.__class__.__name__ + f'(ratio={self.ratio})'
        return format_string


import lmdb
import pyarrow as pa
from torch.nn import Module
import torch.utils.data as data
import os
import torch
import random
random.seed(1234)

from copy import deepcopy

class LMDBDataset(data.Dataset):
    def __init__(self, db_path,split, subset=None, transform=None, target_transform=None):
        self.db_path = db_path
        lmdb_path = None
        if split=="train":
            lmdb_path = os.path.join(self.db_path,"train.lmdb")
        elif split=="valid":
            lmdb_path = os.path.join(self.db_path,"valid.lmdb")
        else:
            lmdb_path = os.path.join(self.db_path,"eval.lmdb")
        self.subset = subset
        self.env = lmdb.open(lmdb_path, subdir=os.path.isdir(lmdb_path),
                             readonly=True, lock=False,
                             readahead=False, meminit=False)
        with self.env.begin(write=False) as txn:
            # self.length = txn.stat()['entries'] - 1
            
            self.length =pa.deserialize(txn.get(b'__len__'))
            self.keys= pa.deserialize(txn.get(b'__keys__'))
            self.org_keys = deepcopy(self.keys)
            self.start = 0
            if subset is not None and subset < self.length:
                self.length = subset
                random.shuffle(self.keys)
                self.org_keys = deepcopy(self.keys)
                self.keys = self.keys[:subset]
                self.start=subset

        self.transform = transform
        self.target_transform = target_transform
        with self.env.begin(write=False) as txn:
            byteflow = txn.get(self.keys[0])
        unpacked = pa.deserialize(byteflow)
        self.num_classes = unpacked[1].shape[1]
        self.txn = self.env.begin(write=False)

        self.sr = 16000

    def __getitem__(self, index):
        env = self.env

        byteflow = self.txn.get(self.keys[index])
        unpacked = pa.deserialize(byteflow)

        waveform, label = torch.from_numpy(unpacked[0]).squeeze(0),torch.from_numpy(unpacked[1]).squeeze(0)
        length = waveform.shape[-1]
        if length > self.sr * 5:
            length = 501 
        else:
            length = length // 160 + 1

        seg_len = int(self.sr)


        if self.transform is not None:
            transformed = self.transform(waveform)
            if self.target_transform is not None:
                transformed = list(transformed)
                transformed[0],label = self.target_transform(transformed[0],label)
                transformed = tuple(transformed)
        
            return transformed, label 
        else:
            return waveform, label
    def cycle(self):
        if self.start + self.subset > len(self.org_keys):
            self.keys = self.org_keys[self.start:] + self.org_keys[:self.start+self.subset - len(self.org_keys)]
            random.shuffle(self.org_keys)
            self.start = 0
            #self.start = self.start + self.subset - len(self.org_keys)
        else:
            self.keys = self.org_keys[self.start:self.start+self.subset]
            self.start = self.start+self.subset

    def __len__(self):
        return len(self.keys)

    def __repr__(self):
        return self.__class__.__name__ + ' (' + self.db_path + ')'




import torch
from torch import nn
from audiossl.modules.transformer import Block
from torch.nn import functional as F
from functools import partial
import time
import warnings
import math

def _no_grad_trunc_normal_(tensor, mean, std, a, b):
    # Cut & paste from PyTorch official master until it's in a few official releases - RW
    # Method based on https://people.sc.fsu.edu/~jburkardt/presentations/truncated_normal.pdf
    def norm_cdf(x):
        # Computes standard normal cumulative distribution function
        return (1. + math.erf(x / math.sqrt(2.))) / 2.

    if (mean < a - 2 * std) or (mean > b + 2 * std):
        warnings.warn("mean is more than 2 std from [a, b] in nn.init.trunc_normal_. "
                      "The distribution of values may be incorrect.",
                      stacklevel=2)

    with torch.no_grad():
        # Values are generated by using a truncated uniform distribution and
        # then using the inverse CDF for the normal distribution.
        # Get upper and lower cdf values
        l = norm_cdf((a - mean) / std)
        u = norm_cdf((b - mean) / std)

        # Uniformly fill tensor with values from [l, u], then translate to
        # [2l-1, 2u-1].
        tensor.uniform_(2 * l - 1, 2 * u - 1)

        # Use inverse cdf transform for normal distribution to get truncated
        # standard normal
        tensor.erfinv_()

        # Transform to proper mean, std
        tensor.mul_(std * math.sqrt(2.))
        tensor.add_(mean)

        # Clamp to ensure it's in the proper range
        tensor.clamp_(min=a, max=b)
        return tensor


def trunc_normal_(tensor, mean=0., std=1., a=-2., b=2.):
    # type: (Tensor, float, float, float, float) -> Tensor
    return _no_grad_trunc_normal_(tensor, mean, std, a, b)


def get_num_patches(height=64,width=1001,patch_height=16,patch_width=16):
    return (height // patch_height) * (width // patch_width)

from einops.layers.torch import Rearrange
class PatchEmbed_v2(nn.Module):
    def __init__(self,patch_height=16,patch_width=16,embed_dim=768):
        super().__init__()
        self.patch_height = patch_height
        self.patch_width = patch_width
        self.patch_maker = Rearrange('b c (h p1) (w p2) -> b (w h) (p1 p2 c)', p1 = patch_height, p2 = patch_width)
        self.patch_embed = nn.Linear(patch_height*patch_width,embed_dim)
        
    def forward(self,melspec,length=None):
        height = melspec.shape[2] - melspec.shape[2]%self.patch_height
        width = melspec.shape[3] - melspec.shape[3]%self.patch_width
        patch = self.patch_maker(melspec[:,:,:height,:width])
        patch_embed = self.patch_embed(patch)

        if length is not None:
            patch_length = (height//self.patch_height) * ((length - length%self.patch_width)//self.patch_width)
        else:
            patch_length = None

        return patch,patch_embed,patch_length


class AST(nn.Module):
    """ Vision Transformer """
    def __init__(self,use_cls=True, spec_h=64,spec_w=1001, patch_w=16,patch_h=16, in_chans=1, num_classes=0, embed_dim=768, depth=12,
                 num_heads=12, mlp_ratio=4., qkv_bias=False, qk_scale=None, drop_rate=0., attn_drop_rate=0.,
                 drop_path_rate=0.1, norm_layer=nn.LayerNorm,mask_ratio=0,pos_type="cut", **kwargs):
        super().__init__()
        self.num_features = self.embed_dim = embed_dim
        self.spec_w = spec_w
        self.spec_h = spec_h
        self.embed_dim = embed_dim
        self.patch_w = patch_w
        self.patch_h = patch_h


        self.patch_embed = PatchEmbed_v2(patch_h,patch_w,embed_dim)
        self.mask_embed = nn.Parameter(torch.zeros(1,1, self.embed_dim))

        num_patches = get_num_patches(spec_h,spec_w,patch_h,patch_w)
        self.num_patches = num_patches
        self.use_cls = use_cls

        if use_cls:
            self.cls_token = nn.Parameter(torch.zeros(1, 1, embed_dim))

        self.pos_embed = nn.Parameter(torch.zeros(1, num_patches + 1, embed_dim))
        self.pos_drop = nn.Dropout(p=drop_rate)
        self.mask_ratio = mask_ratio
        self.pos_type = pos_type

        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, depth)]  # stochastic depth decay rule
        self.blocks = nn.ModuleList([
            Block(
                dim=embed_dim, num_heads=num_heads, mlp_ratio=mlp_ratio, qkv_bias=qkv_bias, qk_scale=qk_scale,
                drop=drop_rate, attn_drop=attn_drop_rate, drop_path=dpr[i], norm_layer=norm_layer)
            for i in range(depth)])
        self.norm = norm_layer(embed_dim)
        if self.mask_ratio > 0:
            self.final_to_melsp = nn.Linear(embed_dim,patch_h*patch_w)


        trunc_normal_(self.pos_embed, std=.02)
        trunc_normal_(self.mask_embed, std=.02)
        if use_cls:
            trunc_normal_(self.cls_token, std=.02)
        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)

    def interpolate_pos_encoding(self, x, h, w):
        npatch = x.shape[1] - 1
        N = self.pos_embed.shape[1] - 1
        if npatch == N and w == self.spec_w and h == self.spec_h:
            return self.pos_embed
        class_pos_embed = self.pos_embed[:, 0]
        patch_pos_embed = self.pos_embed[:, 1:]
        dim = x.shape[-1]
        w0 = w // self.patch_embed.patch_width
        h0 = h // self.patch_embed.patch_height
        # we add a small number to avoid floating point error in the interpolation
        # see discussion at https://github.com/facebookresearch/dino/issues/8
        w0, h0 = w0 + 0.1, h0 + 0.1
        patch_pos_embed = nn.functional.interpolate(
            patch_pos_embed.reshape(1, self.spec_h//self.patch_h, self.spec_w//self.patch_w, dim).permute(0, 3, 1, 2),
            scale_factor=(h0 / (self.spec_h//self.patch_h), w0 / (self.spec_w//self.patch_w)),
            mode='bicubic',
        )
        assert int(h0) == patch_pos_embed.shape[-2] and int(w0) == patch_pos_embed.shape[-1]
        patch_pos_embed = patch_pos_embed.permute(0, 2, 3, 1).view(1, -1, dim)
        return torch.cat((class_pos_embed.unsqueeze(0), patch_pos_embed), dim=1)

    def prepare_tokens(self, x, mask_index, length, mask=True):
        B, nc, h, w = x.shape
        mel_patches,x,patch_length = self.patch_embed(x,length)  # patch linear embedding
        B, T, C = x.shape

        if (mask_index is not None) and mask:
            mask_index_expand = mask_index.unsqueeze(2).expand(B,T,self.embed_dim).float()
            x = (1-mask_index_expand) * x + mask_index_expand * self.mask_embed.expand(B,T,C)

        # add the [CLS] token to the embed patch tokens
        if self.use_cls:
            cls_tokens = self.cls_token.expand(B, -1, -1)
            x = torch.cat((cls_tokens, x), dim=1)


            # add positional encoding to each token

            if self.pos_type == "cut":
                pos = self.pos_embed[:,:T+1,:].expand(B,-1,-1) 
                x = x + pos
            else:
                if patch_length is not None:
                    assert all([patch_length[i] + 1 ==  x.shape[1] for i in range(B)]) or print(x.shape,patch_length)
                pos = self.interpolate_pos_encoding(x,h,w)
                x = x + pos
        else:
            if self.pos_type == "cut":
                pos = self.pos_embed[:,:T,:].expand(B,-1,-1) 
                x = x + pos
            else:
                print("don't support position type interplation when use_cls is set to be False")
                exit()

        return self.pos_drop(x),pos,mel_patches,h,w,patch_length

    def forward(self, x, mask_index=None,length=None,avg=False):
        x,pos,mel_patches,h,w,patch_length = self.prepare_tokens(x,mask_index,length)

        avg_x = []
        for i,blk in enumerate(self.blocks):
            if self.use_cls:
                x = blk(x,patch_length+1)
            else:
                x = blk(x,patch_length)
            if avg:
                if i> len(self.blocks)-9 :
                    avg_x.append(x)
                    #avg_x.append(F.instance_norm(x.transpose(1,2)).transpose(1,2))
        x = self.norm(x)
        if avg:
            avg_x=torch.mean(torch.stack(avg_x),dim=0)

        if self.mask_ratio>0:
            if patch_length is not None:
                length_mask = torch.arange(mel_patches.shape[1]).to(x.device) < patch_length.unsqueeze(1)
                length_mask = length_mask.to(x.device)
                mask = mask_index & length_mask
            else:
                mask = mask_index
                
            logits = self.final_to_melsp(x[:,1:][mask])
            mse_loss = F.mse_loss(logits,mel_patches[mask])
        else:
            mse_loss = torch.zeros([])
        
        if self.use_cls:
            if avg:
                return avg_x[:,0]
            else:
                return x[:,0]
        else:
            if patch_length is not None:
                length_mask = torch.arange(mel_patches.shape[1]).to(x.device) < patch_length.unsqueeze(1)
                length_mask = length_mask.to(x.device)

            if avg:
                avg_t = torch.sum(avg_x*length_mask.unsqueeze(2),dim=1)/patch_length.unsqueeze(1)
                return avg_t
            else:
                avg_t = torch.sum(x*length_mask.unsqueeze(2),dim=1)/patch_length.unsqueeze(1)
                return avg_t

    def get_last_selfattention(self, x):
        x,_,_,_,_,_ = self.prepare_tokens(x,mask_index=None,length=None,mask=False)
        for i, blk in enumerate(self.blocks):
            if i < len(self.blocks) - 1:
                x = blk(x)
            else:
                # return attention of the last block
                return blk(x, return_attention=True)
    def get_intermediate_layers(self, x,length, n=1):
        x,_,_,_,_,patch_length = self.prepare_tokens(x,mask_index=None,length=length,mask=False)
        # we return the output tokens from the `n` last blocks
        output = []
        for i, blk in enumerate(self.blocks):
            if patch_length is None:
                x = blk(x,length=None)
                if self.use_cls:
                    x_ = x
                else:
                    x_ = torch.cat([torch.zeros_like(x[:,0:1]),x],dim=1)
            else:
                if self.use_cls:
                    x = blk(x,length=patch_length+1)
                    x_ = x
                else:
                    x = blk(x,length=patch_length)
                    x_ = torch.cat([torch.zeros_like(x[:,0:1]),x],dim=1)
            if len(self.blocks) - i <= n:
                output.append(self.norm(x_))
        return output
        
    def get_intermediate_layers_chunks(self, x,length, n=1,  chunk_len=401, avgpool=True):
        total_len = x.shape[-1]
        num_chunks = total_len // chunk_len + 1
        cls = []
        avg = []
        chunk_mark = []
        for i in range(num_chunks):
            cur_len = torch.clip(length - i*chunk_len,0)
            if i==0:
                chunk_mark_ = cur_len > 0
            else:
                chunk_mark_ = cur_len > chunk_len//2
            start = i*chunk_len
            end = (i+1) * chunk_len
            if end > total_len:
                end = total_len
            x_chunk=x[:,:,:,start:end]
            x_chunk,_,_,_,_,patch_length = self.prepare_tokens(x_chunk,mask_index=None,length=cur_len,mask=False)
            # we return the output tokens from the `n` last blocks
            output_i = []
            for j, blk in enumerate(self.blocks):
                if patch_length is None:
                    x_chunk = blk(x_chunk,length=None)
                else:
                    if self.use_cls:
                        x_chunk= blk(x_chunk,length=patch_length+1)
                    else:
                        x_chunk= blk(x_chunk,length=patch_length)
                if len(self.blocks) - j <= n:
                    output_i.append(self.norm(x_chunk))
            cls_,avg_=get_cls_avg(output_i,patch_length,self.use_cls)
            """
            print("{}-th chunk".format(i))
            print(len(cls_),cls_[0].shape)
            print(len(avg_),avg_[0].shape)
            print(length)
            print(cur_len)
            print(chunk_mark_.shape)
            print("{}-th chunk".format(i))
            """
            cls.append(cls_)
            avg.append(avg_)
            chunk_mark.append([chunk_mark_]*n)
        cls = [list(x) for x in zip(*cls)]
        avg = [list(x) for x in zip(*avg)]
        chunk_mark = [list(x) for x in zip(*chunk_mark)]
        """
        print(len(cls))
        print(type(cls[0]))
        print(len(cls[0]))
        """

        cls_out = []
        avg_out = []
        for cls_,avg_,chunk_mark_ in zip(cls,avg,chunk_mark):
            """
            print(type(cls_))
            print(type(avg_))
            print(type(chunk_mark_))
            print(len(cls_))
            print(len(avg_))
            print(len(chunk_mark_))
            print(type(cls_[0]))
            print(type(avg_[0]))
            print(type(chunk_mark_[0]))
            print(cls_[0].shape)
            print(avg_[0].shape)
            print(chunk_mark_[0].shape)
            """
            cls_ = torch.stack(cls_,dim=0)
            avg_ = torch.stack(avg_,dim=0)
            chunk_mark_ = torch.stack(chunk_mark_,dim=0)
            """
            print(cls_.shape)
            print(avg_.shape)
            print(cls_[0][0])
            print(cls_[1][0])
            print(avg_[0][0])
            print(avg_[1][0])
            print(chunk_mark_.shape)
            """
            cls_=torch.sum(cls_*chunk_mark_.unsqueeze(-1),dim=0)/torch.sum(chunk_mark_.unsqueeze(-1),dim=0)
            avg_=torch.sum(avg_*chunk_mark_.unsqueeze(-1),dim=0)/torch.sum(chunk_mark_.unsqueeze(-1),dim=0)
            """
            print(cls_[0])
            print(avg_[0])
            print(cls_.shape)
            print(avg_.shape)
            """

            cls_out.append(cls_)
            avg_out.append(avg_)
        if avgpool:
            return torch.cat(cls_out+avg_out,dim=-1)
        else:
            return torch.cat(cls_out,dim=-1)
        #return cls_out,avg_out

def get_cls_avg(output_i,cur_len,use_cls):
    if use_cls:
        length_mask = torch.arange(output_i[0].shape[1]-1).to(output_i[0].device) < cur_len.unsqueeze(1)
    else:
        length_mask = torch.arange(output_i[0].shape[1]).to(output_i[0].device) < cur_len.unsqueeze(1)
    if use_cls:
        cls = [x[:,0] for x in output_i]
        avg = [torch.sum(x[:,1:]*length_mask.unsqueeze(-1),dim=1)/(cur_len.unsqueeze(1)+1e-6) for x in output_i]
    else:
        cls = [torch.zeros_like(x[:,0]) for x in output_i]
        avg = [torch.sum(x*length_mask.unsqueeze(-1),dim=1)/(cur_len.unsqueeze(1)+1e-6) for x in output_i]
    return cls,avg
def AST_small(patch_h=64,patch_w=4,**kwargs):
    return AST(patch_h=patch_h,patch_w=patch_w,embed_dim=384,depth=12,num_heads=6,qkv_bias=False,norm_layer=partial(nn.LayerNorm, eps=1e-6),**kwargs)

def AST_base(patch_h=64,patch_w=4,**kwargs):
    return AST(patch_h=patch_h,patch_w=patch_w,embed_dim=768,depth=12,num_heads=12,qkv_bias=False,norm_layer=partial(nn.LayerNorm, eps=1e-6),**kwargs)

def AST_large(patch_h,patch_w,**kwargs):
    return AST(patch_h=patch_h,patch_w=patch_w,embed_dim=1024,depth=24,num_heads=16,qkv_bias=False,norm_layer=partial(nn.LayerNorm, eps=1e-6),**kwargs)





from torch import nn
from audiossl.models.atst.byol import MultiCropWrapper,ByolLoss
from audiossl.models.atst.audio_transformer import AST_small,AST_base
import torch
class ATST(nn.Module):
    def __init__(self,arch="small",ncrops=2,**kwargs):
        super().__init__()
        if arch == "small":
            encoder_fn = AST_small
            embed_dim = 384
        elif arch == "base":
            encoder_fn = AST_base
            embed_dim = 768
        else:
            raise RuntimeError("arch {} is not implemented".format(arch))
        self.student=MultiCropWrapper(encoder_fn(**kwargs),embed_dim,predictor=True)
        self.teacher=MultiCropWrapper(encoder_fn(**kwargs),embed_dim,predictor=False)
        for p in self.teacher.parameters():
            p.requires_grad = False
        self.teacher.load_state_dict({k:v for k,v in self.student.state_dict().items() if "predictor" not in k })
        self.loss_fn = ByolLoss(ncrops)
    def forward(self,melspecs,lengths):
        teacher_output = self.teacher(melspecs[:2],lengths[:2])  
        student_output = self.student(melspecs,lengths)
        loss = self.loss_fn(student_output,teacher_output)
        return loss
    def update_teacher(self,m):
        with torch.no_grad():
            for param_q, param_k in zip(self.student.encoder.parameters(), self.teacher.encoder.parameters()):
                param_k.data.mul_(m).add_((1 - m) * param_q.detach().data)
            for param_q, param_k in zip(self.student.projector.parameters(), self.teacher.projector.parameters()):
                param_k.data.mul_(m).add_((1 - m) * param_q.detach().data)

        


import torch
import numpy as np
@torch.no_grad()
def concat_all_gather(tensor):
    """
    Performs all_gather operation on the provided tensors.
    *** Warning ***: torch.distributed.all_gather has no gradient.
    """
    tensors_gather = [torch.ones_like(tensor)
        for _ in range(torch.distributed.get_world_size())]
    torch.distributed.all_gather(tensors_gather, tensor, async_op=False)

    output = torch.cat(tensors_gather, dim=0)
    return output

def cosine_scheduler_epoch(base_value, final_value, epochs, niter_per_ep, warmup_epochs=0, start_warmup_value=0):
    warmup_schedule = np.array([])
    warmup_iters = warmup_epochs * niter_per_ep
    if warmup_epochs > 0:
        warmup_schedule = np.linspace(start_warmup_value, base_value, warmup_iters)

    iters = np.arange(epochs * niter_per_ep - warmup_iters)
    schedule = final_value + 0.5 * (base_value - final_value) * (1 + np.cos(np.pi * iters / len(iters)))

    schedule = np.concatenate((warmup_schedule, schedule))
    assert len(schedule) == epochs * niter_per_ep
    return schedule

def cosine_scheduler_step(base_value, final_value, max_steps, warmup_steps=0, start_warmup_value=0):
    warmup_schedule = np.array([])
    if warmup_steps > 0:
        warmup_schedule = np.linspace(start_warmup_value, base_value, warmup_steps)

    iters = np.arange(max_steps - warmup_steps)
    schedule = final_value + 0.5 * (base_value - final_value) * (1 + np.cos(np.pi * iters / len(iters)))

    schedule = np.concatenate((warmup_schedule, schedule))
    assert len(schedule) == max_steps
    return schedule

def get_params_groups(model):
    regularized = []
    not_regularized = []
    for name, param in model.named_parameters():
        if not param.requires_grad:
            continue
        # we do not regularize biases nor Norm parameters
        if name.endswith(".bias") or len(param.shape) == 1:
            not_regularized.append(param)
        else:
            regularized.append(param)
    return [{'params': regularized}, {'params': not_regularized, 'weight_decay': 0.}]
def bool_flag(s):
    """
    Parse boolean arguments from the command line.
    """
    FALSY_STRINGS = {"off", "false", "0"}
    TRUTHY_STRINGS = {"on", "true", "1"}
    if s.lower() in FALSY_STRINGS:
        return False
    elif s.lower() in TRUTHY_STRINGS:
        return True
    else:
        raise argparse.ArgumentTypeError("invalid value for a boolean flag")



import torch
import numpy as np
class MixupSpecLabel:
    def __init__(self,alpha=0.5,n_memory=5000,num_classes=100):
        self.alpha = alpha
        self.memory_bank = []
        self.n=n_memory
        self.num_classes=num_classes
    def __call__(self,x,y):

        """convert label to one hot vector"""
        if isinstance(y,int) or (isinstance(y,torch.Tensor) and (len(y.shape)==0 or y.shape[-1]==1) ):
            if isinstance(y,int):
                y=torch.nn.functional.one_hot(torch.tensor(y),num_classes=self.num_classes)
            else:
                y=torch.nn.functional.one_hot(y.to(torch.int64),num_classes=self.num_classes)

        l = np.random.beta(self.alpha,self.alpha,1)

        if self.memory_bank:
            x_,y_ = self.memory_bank[np.random.randint(len(self.memory_bank))]
            #x_mix = torch.log(x.exp()*l +x_.exp()*energy_scale(x,x_) * (1-l) + torch.finfo(x.dtype).eps)
            if x.shape[-1] == x_.shape[-1]:
                x_mix = x*l + x_*(1-l)
            elif x.shape[-1] > x_.shape[-1]:
                start = np.random.randint(0,x.shape[-1] - x_.shape[-1])
                x_mix = x.clone()

                x_mix[:,:,start:start+x_.shape[-1]] = x[:,:,start:start+x_.shape[-1]]*l + x_*(1-l)
            else:
                start = np.random.randint(0,x_.shape[-1] - x.shape[-1])

                x_mix= x*l + x_[:,:,start:start+x.shape[-1]]*(1-l)
            y_mix = y*l +y_ * (1-l)
        else:
            x_mix=x
            y_mix=y

        self.memory_bank.append((x,y))
        self.memory_bank = self.memory_bank[-self.n:]
        return x_mix.to(torch.float),y_mix.to(torch.float)



import torch
import torch.nn as nn
from torch.nn import functional as F

def build_mlp(num_layers, input_dim, mlp_dim, output_dim, last_bn=True):
    mlp = []
    for l in range(num_layers):
        dim1 = input_dim if l == 0 else mlp_dim
        dim2 = output_dim if l == num_layers - 1 else mlp_dim

        mlp.append(nn.Linear(dim1, dim2, bias=False))

        if l < num_layers - 1:
            mlp.append(nn.BatchNorm1d(dim2))
            mlp.append(nn.ReLU(inplace=True))
        elif last_bn:
            # follow SimCLR's design: https://github.com/google-research/simclr/blob/master/model_util.py#L157
            # for simplicity, we further removed gamma in BN
            mlp.append(nn.BatchNorm1d(dim2, affine=False))

    return nn.Sequential(*mlp)

def byol_loss_func(p: torch.Tensor, z: torch.Tensor, simplified: bool = True) :
    """
    Computes BYOL's loss given batch of predicted features p and projected momentum features z.
    Args:
        p (torch.Tensor): NxD Tensor containing predicted features from view 1
        z (torch.Tensor): NxD Tensor containing projected momentum features from view 2
        simplified (bool): faster computation, but with same result. Defaults to True.
    Returns:
        torch.Tensor: BYOL loss.
    """

    if simplified:
        return 2 - 2 * F.cosine_similarity(p, z, dim=-1).mean()

    p = F.normalize(p, dim=-1)
    z = F.normalize(z, dim=-1)

    return 2 - 2 * (p * z).sum(dim=1).mean()
def compute_var(y):
        y = y.view(-1, y.size(-1))
        zc = torch.tensor(y.size(0)).cuda()
        zs = y.sum(dim=0)
        zss = (y ** 2).sum(dim=0)

        torch.distributed.all_reduce(zc)
        torch.distributed.all_reduce(zs)
        torch.distributed.all_reduce(zss)

        var = zss / (zc - 1) - (zs ** 2) / (zc * (zc - 1))
        return torch.sqrt(var + 1e-6)



class ByolLoss(nn.Module):
    def __init__(self,ncrops):
        super().__init__()
        self.ncrops=ncrops
    def forward(self,student,teacher):
        std_cls_s=compute_var(F.normalize(student,dim=-1)).mean()
        std_cls_t=compute_var(F.normalize(teacher,dim=-1)).mean()
        student = student.chunk(self.ncrops)
        teacher = teacher.detach().chunk(2)
        total_loss = 0
        n_loss_terms = 0
        for iq,q in enumerate(teacher):
            for iv,v in enumerate(student):
                if iq==iv:
                    continue
            
                loss = byol_loss_func(q,v,simplified=False)
                n_loss_terms += 1
                total_loss += loss.mean()

        total_loss /= n_loss_terms
        return total_loss,std_cls_s,std_cls_t



class MultiCropWrapper(nn.Module):
    """
    Perform forward pass separately on each resolution input.
    The inputs corresponding to a single resolution are clubbed and single
    forward is run on the same resolution inputs. Hence we do several
    forward passes = number of different resolutions used. We then
    concatenate all the output features and run the head forward on these
    concatenated features.
    """
    def __init__(self, encoder,
                 embed_dim, 
                 predictor=True):
        super(MultiCropWrapper, self).__init__()
        # disable layers dedicated to ImageNet labels classification
        self.encoder = encoder
        self.projector = build_mlp(2,embed_dim,4096,256,last_bn=False)
        if predictor:
            self.predictor=build_mlp(2,256,4096,256,last_bn=False)
        else: 
            self.predictor=nn.Identity()

    def forward(self, x, length, avg=False):
        # convert to list
        if not isinstance(x, list):
            x = [x]
        idx_crops = torch.cumsum(torch.unique_consecutive(
            torch.tensor([inp.shape[-1] for inp in x]),
            return_counts=True,
        )[1], 0)
        start_idx, output = 0,torch.empty(0).to(x[0].device)

        for end_idx in idx_crops:
            _out = self.encoder(torch.cat(x[start_idx: end_idx]),
                                          length=torch.cat(length[start_idx:end_idx]),
                                          avg=avg)
            # accumulate outputs
            output = torch.cat((output, _out))
            start_idx = end_idx
        # Run the head forward on the concatenated features.
        return self.predictor(self.projector(output))





# -*- coding: utf-8 -*- #
"""copied and modified from s3prl"""
"""
    
    FileName     [ dataset.py ]
    Synopsis     [ the emotion classifier dataset ]
    Copyright    [ Copyleft(c), Speech Lab, NTU, Taiwan ]
"""

import json
from pathlib import Path
from os.path import join as path_join

import torchaudio
from torch.utils.data import Dataset
from torchaudio.transforms import Resample

SAMPLE_RATE = 16000


class IEMOCAPDataset(Dataset):
    def __init__(self, data_dir, meta_path, pre_load=True,transform=None):
        self.data_dir = data_dir
        self.pre_load = pre_load
        self.transform = transform
        with open(meta_path, 'r') as f:
            self.data = json.load(f)
        self.class_dict = self.data['labels']
        self.idx2emotion = {value: key for key, value in self.class_dict.items()}
        self.class_num = len(self.class_dict)
        self.meta_data = self.data['meta_data']
        _, origin_sr = torchaudio.load(
            path_join(self.data_dir, self.meta_data[0]['path']))
        self.resampler = Resample(origin_sr, SAMPLE_RATE)
        if self.pre_load:
            self.wavs = self._load_all()

    def _load_wav(self, path):
        wav, _ = torchaudio.load(path_join(self.data_dir, path))
        wav = self.resampler(wav)
        return wav

    def _load_all(self):
        wavforms = []
        for info in self.meta_data:
            wav = self._load_wav(info['path'])
            wavforms.append(wav)
        return wavforms

    def __getitem__(self, idx):
        label = self.meta_data[idx]['label']
        label = self.class_dict[label]
        if self.pre_load:
            wav = self.wavs[idx]
        else:
            wav = self._load_wav(self.meta_data[idx]['path'])
        if self.transform is None:
            return wav, label
        else:
            return self.transform(wav),label

    def __len__(self):
        return len(self.meta_data)

def collate_fn(samples):
    return zip(*samples)


from setuptools import setup,find_packages


packages=find_packages(where=".")

setup(
    packages=packages,
    entry_points={
                  'console_scripts':[
                                    'atst_train=audiossl.methods.atst.train:main' ,
                                    'atst_downstream_train_freeze=audiossl.methods.atst.downstream.train_freeze:main' ,
                                    'atst_downstream_train_finetune=audiossl.methods.atst.downstream.train_finetune:main' 
                                    ]
                 }
      )




"""copied and modified from s3prl"""
import torch
from torch.utils.data import DataLoader, Dataset
import numpy as np 
from librosa.util import find_files
from torchaudio import load
from torch import nn
import os 
import re
import random
import pickle
import torchaudio
import sys
import time
import glob
import tqdm
from pathlib import Path
from audiossl.datasets import register_dataset

CACHE_PATH = os.path.join(os.path.dirname(__file__), '.cache/')


# Voxceleb 1 Speaker Identification

class SpeakerClassifiDataset(Dataset):
    def __init__(self, mode, file_path, meta_data, max_timestep=None,sr=16000, transform=None, target_transform=None):

        self.root = file_path
        self.speaker_num = 1251
        self.meta_data =meta_data
        self.max_timestep = max_timestep
        self.sr=sr
        self.transform = transform
        self.target_transform = target_transform
        self.usage_list = open(self.meta_data, "r").readlines()

        cache_path = os.path.join(CACHE_PATH, f'{mode}.pkl')
        if os.path.isfile(cache_path):
            print(f'[SpeakerClassifiDataset] - Loading file paths from {cache_path}')
            with open(cache_path, 'rb') as cache:
                dataset = pickle.load(cache)
        else:
            dataset = eval("self.{}".format(mode))()
            os.makedirs(os.path.dirname(cache_path), exist_ok=True)
            with open(cache_path, 'wb') as cache:
                pickle.dump(dataset, cache)
        print(f'[SpeakerClassifiDataset] - there are {len(dataset)} files found')

        self.dataset = dataset
        self.label = self.build_label(self.dataset)

    # file_path/id0001/asfsafs/xxx.wav
    def build_label(self, train_path_list):

        y = []
        for path in train_path_list:
            id_string = path.split("/")[-3]
            y.append(int(id_string[2:]) - 10001)

        return y
    
    @classmethod
    def label2speaker(self, labels):
        return [f"id{label + 10001}" for label in labels]
    
    def train(self):

        dataset = []
        print("search specified wav name for training set")
        for string in tqdm.tqdm(self.usage_list):
            pair = string.split()
            index = pair[0]
            if int(index) == 1:
                x = list(self.root.glob("*/wav/" + pair[1]))
                dataset.append(str(x[0]))
        print("finish searching training set wav")
                
        return dataset
        
    def dev(self):

        dataset = []
        print("search specified wav name for dev set")
        for string in tqdm.tqdm(self.usage_list):
            pair = string.split()
            index = pair[0]
            if int(index) == 2:
                x = list(self.root.glob("*/wav/" + pair[1]))
                dataset.append(str(x[0])) 
        print("finish searching dev set wav")

        return dataset       

    def test(self):

        dataset = []
        print("search specified wav name for test set")
        for string in tqdm.tqdm(self.usage_list):
            pair = string.split()
            index = pair[0]
            if int(index) == 3:
                x = list(self.root.glob("*/wav/" + pair[1]))
                dataset.append(str(x[0])) 
        print("finish searching test set wav")

        return dataset

    def __len__(self):
        return len(self.dataset)
    
    def __getitem__(self, idx):
        wav, sr = torchaudio.load(self.dataset[idx])
        if not (self.sr == sr):
            raise "sampling rate {} is expected, while {} is given".format(
                self.sr, sr)
        length = wav.shape[0]

        if self.max_timestep !=None:
            if length > self.max_timestep:
                start = random.randint(0, int(length-self.max_timestep))
                wav = wav[start:start+self.max_timestep]
                length = self.max_timestep

        def path2name(path):
            return Path("-".join((Path(path).parts)[-3:])).stem

        path = self.dataset[idx]

        if self.transform is None:

            return wav, self.label[idx]
        else:
            wav = self.transform(wav)
            label = self.label[idx]
            if self.target_transform is not None:
                wav = list(wav)
                wav[0],label=self.target_transform(wav[0],label)
                wav = tuple(wav)
            return wav,label

        
    def collate_fn(self, samples):
        return zip(*samples)


from matplotlib import pyplot as plt
def plot_spec(x,save_path):
    t = range(0,x.shape[0])
    f = range(0,x.shape[1])
    plt.pcolormesh(x)
    plt.xlabel('frequency',fontsize=20)
    plt.ylabel('time',fontsize=20)
    plt.savefig(save_path)
    plt.close()




